/*
 * Copyright (C) 2005-2012 Alfresco Software Limited.
 *
 * This file is part of Alfresco
 *
 * Alfresco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Alfresco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Alfresco. If not, see <http://www.gnu.org/licenses/>.
 */
package org.alfresco.bm.classloader.zookeeper;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.I0Itec.zkclient.IZkChildListener;
import org.I0Itec.zkclient.IZkDataListener;
import org.I0Itec.zkclient.ZkClient;
import org.I0Itec.zkclient.exception.ZkMarshallingError;
import org.alfresco.bm.classloader.CallbackClassLoader;
import org.alfresco.bm.classloader.ClassLoaderListener;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * A Java {@link ClassLoader} pulling classes from ZooKeeper.
 * <p/>
 * This class loader allows listeners to be registered to receive callbacks when
 * the ZooKeeper data changes.  
 * 
 * @author Derek Hulley
 * @since 1.0
 */
public class ZooKeeperClassLoader extends ClassLoader implements CallbackClassLoader, IZkDataListener, IZkChildListener
{
    private static Log logger = LogFactory.getLog(ZooKeeperClassLoader.class);
    
    private final ZkClient zk;
    private final String rootPath;
    private final List<ClassLoaderListener> listeners;
    private final Map<String, Class<?>> foundClasses;
    
    /**
     * @param parentClassloader         the (required) parent class loader
     * @param zk                        the ZK client to run against
     * @param rootPath                  the ZK root path (akin to a classpath entry)
     */
    public ZooKeeperClassLoader(ClassLoader parentClassloader, ZkClient zk, String rootPath)
    {
        super(parentClassloader);
        
        assert(parentClassloader != null);
        assert(zk != null);
        assert(rootPath != null);
        assert(rootPath.length() > 0) : "ZooKeeper root path length is zero.";
        assert(rootPath.startsWith("/")) : "ZooKeeper root path must start with '/'.";
        assert(!rootPath.endsWith("/")) : "ZooKeeper root path must not end with '/'.";
        
        this.zk= zk;
        this.rootPath = rootPath;
        this.listeners = new ArrayList<ClassLoaderListener>(2);
        this.foundClasses = new HashMap<String, Class<?>>(13);
    }
    
    /**
     * Calls any listeners.
     */
    @Override
    public void handleChildChange(String parentPath, List<String> currentChilds) throws Exception
    {
        callListeners();
    }

    /**
     * Calls any listeners.
     */
    @Override
    public void handleDataChange(String dataPath, Object data) throws Exception
    {
        callListeners();
    }

    /**
     * Calls any listeners.
     */
    @Override
    public void handleDataDeleted(String dataPath) throws Exception
    {
        callListeners();
    }
    
    /**
     * Call all the associated listeners
     */
    private synchronized void callListeners()
    {
        for (ClassLoaderListener listener : this.listeners)
        {
            listener.classesHaveChanged();
        }
    }

    /**
     * Add a change listener
     * 
     * @param listener          the listener that will be called if there are <b>any</b>
     *                          changes to classes already loaded
     */
    @Override
    public synchronized void addListener(ClassLoaderListener listener)
    {
        // We register for general changes immediately on the first listener
        if (this.listeners.size() == 0)
        {
            zk.subscribeChildChanges(rootPath, this);
        }
        // Add to the list that will be called in event of any change
        this.listeners.add(listener);
    }

    /**
     * Takes the {@link #ZooKeeperClassLoader(ClassLoader, ZkClient, String) rootPath} variable and
     * finds (using '/' in place of '.') the class in
     * {@link #ZooKeeperClassLoader(ClassLoader, ZkClient, String) ZooKeeper}.
     */
    @Override
    public synchronized Class<?> loadClass(String name) throws ClassNotFoundException
    {
        if (foundClasses.containsKey(name))
        {
            return foundClasses.get(name);
        }
        // Replace '.' with '/'
        String pathExt = name.replace('.', '/');
        // Append it to the root path
        String fullPath = rootPath + "/" + pathExt;
        
        // Load the class
        Class<?> clazz = null;
        try
        {
            byte[] bytes = zk.readData(fullPath, true);
            // The class being requested might not be present
            if (bytes != null)
            {
                // Turn into a class
                clazz = this.defineClass(name, bytes, 0, bytes.length);
            }
        }
        catch (Throwable e)
        {
            throw new ZkMarshallingError(
                    "Unable to load class data: \n" +
                    "   Root path: " + rootPath + "\n" +
                    "   Class:     " + fullPath,
                    e);
        }
        if (clazz == null)
        {
            clazz = super.loadClass(name);
            if (clazz == null)
            {
                throw new ClassNotFoundException("Class not found in ZooKeeper at '" + fullPath + "'");
            }
            else
            {
                if (logger.isDebugEnabled())
                {
                    logger.debug("Found class in parent classloader: " + name);
                }
            }
        }
        else
        {
            if (logger.isDebugEnabled())
            {
                logger.info("Found class in ZooKeeper at '" + fullPath + "'");
            }
            // Subscribe to listen for any changes to the class
            zk.subscribeDataChanges(fullPath, this);
            // Keep it for later requests
            foundClasses.put(name, clazz);
            // Register a listener to ensure that we are notified of any changes or removals
            zk.subscribeDataChanges(fullPath, this);
        }
        // Done
        return clazz;
    }
}
