/*
 * Copyright (C) 2005-2012 Alfresco Software Limited.
 *
 * This file is part of Alfresco
 *
 * Alfresco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Alfresco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Alfresco. If not, see <http://www.gnu.org/licenses/>.
 */
package org.alfresco.bm.config;

import java.net.URLClassLoader;
import java.util.Properties;
import java.util.Set;

/**
 * Interface for classes that are able to provide configuration specifically for 
 * the benchmark server.
 * 
 * @author Derek Hulley
 * @since 1.0
 */
public interface ConfigService extends ConfigConstants
{
    /**
     * Check if a given path exists
     * 
     * @param path              the path (over and above any root path) to check
     * @return                  <tt>true</tt> if the path exists
     */
    boolean exists(String ... path);
    
    /**
     * Retrieve string data from a given path
     * 
     * @param listener          a listener to call if the data changes (<tt>null</tt> allowed)
     * @param path              the path to the data
     * @return                  Returns the data or <tt>null</tt> if not found
     */
    String getString(ConfigDataListener listener, String ... path);
    
    /**
     * Retrieve the properties from a given path
     * 
     * @param listener          a listener to call if the properties change (<tt>null</tt> allowed)
     * @param path              the path to the properties
     * @return                  Java properties from the location (never <tt>null</tt>)
     */
    Properties getProperties(ConfigDataListener listener, String ... path);
    
    /**
     * Set properties against a path
     * 
     * @param listener          a listener to call if the properties change (<tt>null</tt> allowed)
     * @param ephemeral         <tt>true</tt> if the path should be removed when the server stops
     * @param sequential        <tt>true</tt> if the last path element must have an auto-incrementing
     *                          value attached i.e. if the final path element is always new
     * @param properties        the properties to set
     * @param path              the path to the properties
     * @return                  the final path element, which will be different from the one supplied
     *                          if it is auto-incrementing
     */
    String setProperties(
            ConfigDataListener listener,
            boolean ephemeral, boolean sequential,
            Properties properties,
            String ... path);
    
    /**
     * Retrieve the children of a given path
     * 
     * @param listener          a listener to call if the the children change (<tt>null</tt> allowed)
     * @param path              the path to the parent
     * @return                  Returns the current children of the provided path (never <tt>null</tt>)
     */
    Set<String> getChildren(ConfigChildListener listener, String ... path);
    
    /**
     * Retrieve a {@link ClassLoader} rooted at the given location
     * 
     * @param listener          a listener to call if the classes change (<tt>null</tt> allowed)
     * @param parent            the parent {@link ClassLoader}
     * @param path              the path to the classes root (akin to a classpath)
     * @return                  Returns the new {@link ClassLoader}
     */
    ClassLoader getClassLoader(ConfigClassesListener listener, ClassLoader parent, String ... path);
    
    /**
     * Retrieve a {@link URLClassLoader} rooted at the given location.  The children of the given
     * path are found and each is treated as a separate jar file.
     * 
     * @param listener          a listener to call if the classes change (<tt>null</tt> allowed)
     * @param parent            the parent {@link ClassLoader}
     * @param path              the path to the jar
     * @return                  Returns the new {@link ClassLoader}
     */
    ClassLoader getJarLoader(ConfigClassesListener listener, ClassLoader parent, String ... path);
}
