/*
 * Copyright (C) 2005-2012 Alfresco Software Limited.
 *
 * This file is part of Alfresco
 *
 * Alfresco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Alfresco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Alfresco. If not, see <http://www.gnu.org/licenses/>.
 */
package org.alfresco.bm.config.zookeeper;

import java.util.Properties;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.google.gson.Gson;
import com.google.gson.JsonSyntaxException;

/**
 * Simple container to load node from a ZooKeeper node and convert it to
 * the required target type or {@link java.util.Properties properties} if the
 * java type is not specified.
 * 
 * @author Derek Hulley
 * @since 1.0
 */
public class ZkConfigNode
{
    private static final Log logger = LogFactory.getLog(ZkConfigNode.class);
    
    private static Gson gson = new Gson();
    
    private final String json;
    private final Object instance;

    public static Object getInstance(String json)
    {
        if (logger.isDebugEnabled())
        {
            logger.debug("\n" +
            		"Getting object from JSON: \n" + json);
        }
        Object obj = null;
        
        // First see if there is a class definition in it somewhere
        int classIndex = json.indexOf("class:");
        if (classIndex < 0)
        {
            // No there is not.  Therefore assume it is an instance of Properties
            obj = gson.fromJson(json, Properties.class);
        }
        else
        {
            // Find first '
            int openingQuoteIndex = json.indexOf("'", classIndex + 6);
            if (openingQuoteIndex < 0)
            {
                throw new JsonSyntaxException(
                        "Unable to extract class name: \n" +
                        "   json: " + json);
            }
            // Find the next '
            int closingQuoteIndex = json.indexOf("'", openingQuoteIndex + 1);
            if (closingQuoteIndex < 0)
            {
                throw new JsonSyntaxException(
                        "Unable to extract class name: \n" +
                        "   json: " + json);
            }
            String className = json.substring(openingQuoteIndex + 1, closingQuoteIndex);
            Class<?> clazz = null;
            try
            {
                clazz = Class.forName(className);
            }
            catch (Throwable e)
            {
                throw new JsonSyntaxException(
                        "Unable to load class: " + className + "\n" +
                        "   class: " + className + "\n" +
                        "   json:  " + json,
                        e);
            }
            
            // No make sure we move past any subsequent ','
            int afterClassCommaIndex = json.indexOf(",", closingQuoteIndex + 1);
            if (afterClassCommaIndex < 0)
            {
                throw new JsonSyntaxException(
                        "Expected 'class' value to be first followed by subsequent values for the instance: \n" +
                        "   json: " + json);
            }
            
            // Build new JSON
            StringBuilder newJson = new StringBuilder(json.length());
            newJson.append(json.substring(0, classIndex));
            newJson.append(json.substring(afterClassCommaIndex + 1));
            // We have the new JSON
            obj = gson.fromJson(newJson.toString(), clazz);
        }
        
        // Done
        if (logger.isDebugEnabled())
        {
            logger.debug("\n" +
                    "   JSON converted to : \n" +
                    "       class:    " + obj.getClass().getName() + "\n" +
                    "       instance: " + obj);
        }
        return obj;
    }
    
    /**
     * @param json                      the JSON string
     * @throws JsonSyntaxException      if the JSON could not be converted to a Java instance
     */
    public ZkConfigNode(String json)
    {
        assert(json != null);
        assert(json.length() > 0) : "JSON config string length is zero.";
        
        this.json = json;
        this.instance = getInstance(json);
    }

    @Override
    public String toString()
    {
        return "ZkConfigNode \n" +
        		"[ \n" +
                "   json=" + json + "\n" +
                "   instance=" + instance + "\n" +
                "]";
    }

    /**
     * Get the original JSON from which the {@link #getInstance()} was constructed.
     */
    public String getJson()
    {
        return json;
    }

    /**
     * Get the object made from the supplied JSON
     */
    public Object getInstance()
    {
        return instance;
    }
}
