/*
 * Copyright (C) 2005-2012 Alfresco Software Limited.
 *
 * This file is part of Alfresco
 *
 * Alfresco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Alfresco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Alfresco. If not, see <http://www.gnu.org/licenses/>.
 */
package org.alfresco.bm.config.zookeeper;

import java.io.File;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.I0Itec.zkclient.IZkChildListener;
import org.I0Itec.zkclient.IZkDataListener;
import org.I0Itec.zkclient.ZkClient;
import org.I0Itec.zkclient.exception.ZkNoNodeException;
import org.alfresco.bm.classloader.ClassLoaderListener;
import org.alfresco.bm.classloader.zookeeper.ZooKeeperClassLoader;
import org.alfresco.bm.config.ConfigChildListener;
import org.alfresco.bm.config.ConfigClassesListener;
import org.alfresco.bm.config.ConfigDataListener;
import org.alfresco.bm.config.ConfigException;
import org.alfresco.bm.config.ConfigService;
import org.apache.commons.io.FileUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.json.simple.JSONValue;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.context.ApplicationListener;
import org.springframework.context.event.ApplicationContextEvent;
import org.springframework.context.event.ContextStartedEvent;
import org.springframework.context.event.ContextStoppedEvent;

/**
 * ZooKeeper-based configuration loader that, starting from a root path,
 * loads benchmark-related configurations.
 * 
 * @author Derek Hulley
 * @since 1.0
 */
public class ZkConfigService
        implements
            ConfigService, ApplicationListener<ApplicationContextEvent>, ApplicationContextAware
{
    private static final Log logger = LogFactory.getLog(ZkConfigService.class);
    
    private final ZkClient zk;
    private final String zkPath;
    private ApplicationContext ctx;
    
    /**
     * Construct the configuration loader.
     * 
     * @param zk                the ZooKeeper client
     * @param zkPath            the root path to use in ZK
     */
    public ZkConfigService(ZkClient zk, String zkPath)
    {
        assert(zk != null);
        assert(zkPath != null);
        assert(zkPath.length() > 0) : "ZooKeeper root path length is zero.";
        assert(zkPath.startsWith(SEPARATOR)) : "ZooKeeper root path must start with '/'.";
        assert(!zkPath.endsWith(SEPARATOR)) : "ZooKeeper root path must not end with '/'.";
        
        this.zk = zk;
        this.zkPath = zkPath;
    }

    /**
     * Record the context that started this instance.
     */
    @Override
    public void setApplicationContext(ApplicationContext ctx) throws BeansException
    {
        this.ctx = ctx;
    }

    /**
     * Handles <b>start</b> and <b>stop</b> by checking configuration and cleaning caches respectively.
     */
    @Override
    public void onApplicationEvent(ApplicationContextEvent event)
    {
        // Ignore messages from other contexts
        if (event.getApplicationContext() != this.ctx)
        {
            return;
        }
        if (event instanceof ContextStartedEvent)
        {
            // Started
            startup();
        }
        else if (event instanceof ContextStoppedEvent)
        {
            // Stopped
            shutdown();
        }
    }
    
    /**
     * Nothing to do, yet
     */
    private void startup()
    {
    }
    
    /**
     * Drop any stored configuration and discard any watchers
     */
    public void shutdown()
    {
        // Remove all listeners
        zk.unsubscribeAll();
        zk.close();
    }
    
    /**
     * Combines the {@link ZkConfigService#ZkConfigService(ZkClient, String) root path} with
     * the given path elements
     * 
     * @param path              the path to append to the root path
     * @return                  a full ZK path
     */
    private String makeFullPath(String ... path)
    {
        StringBuilder sb = new StringBuilder(this.zkPath);
        for (String pathElement : path)
        {
            sb.append(SEPARATOR).append(pathElement);
        }
        return sb.toString();
    }
    
    /**
     * Listener mapper to ZooKeeper client interface
     * 
     * @author Derek Hulley
     * @since 1.0
     */
    private static class ZkConfigDataListener implements IZkDataListener
    {
        private final ConfigDataListener listener;
        private ZkConfigDataListener(ConfigDataListener listener)
        {
            this.listener = listener;
        }
        @Override
        public void handleDataChange(String dataPath, Object data) throws Exception
        {
            if (listener != null)
            {
                listener.dataChanged(dataPath, false);
            }
        }
        @Override
        public void handleDataDeleted(String dataPath) throws Exception
        {
            if (listener != null)
            {
                listener.dataChanged(dataPath, true);
            }
        }
    }
    
    /**
     * Listener mapper to ZooKeeper client interface
     * 
     * @author Derek Hulley
     * @since 1.0
     */
    private static class ZkConfigChildListener implements IZkChildListener
    {
        private final ConfigChildListener listener;
        private ZkConfigChildListener(ConfigChildListener listener)
        {
            this.listener = listener;
        }
        @Override
        public void handleChildChange(String parentPath, List<String> currentChilds) throws Exception
        {
            if (listener != null)
            {
                listener.childrenChanged(parentPath);
            }
        }
    }
    
    /**
     * Listener mapper to ZooKeeper class loader
     * 
     * @author Derek Hulley
     * @since 1.0
     */
    private static class ZkConfigClassLoaderListener implements ClassLoaderListener
    {
        private final String path;
        private final ConfigClassesListener listener;
        private ZkConfigClassLoaderListener(String path, ConfigClassesListener listener)
        {
            this.path = path;
            this.listener = listener;
        }
        @Override
        public void classesHaveChanged()
        {
            if (listener != null)
            {
                listener.classesChanged(path);
            }
        }
    }
    
    /**
     * Listener mapper to ZooKeeper class loader
     * 
     * @author Derek Hulley
     * @since 1.0
     */
    private static class ZkConfigJarLoaderListener implements IZkDataListener, IZkChildListener
    {
        private final String path;
        private final ConfigClassesListener listener;
        private ZkConfigJarLoaderListener(String path, ConfigClassesListener listener)
        {
            this.path = path;
            this.listener = listener;
        }
        @Override
        public void handleChildChange(String parentPath, List<String> currentChilds) throws Exception
        {
            classesHaveChanged();
        }
        @Override
        public void handleDataChange(String dataPath, Object data) throws Exception
        {
            classesHaveChanged();
        }
        @Override
        public void handleDataDeleted(String dataPath) throws Exception
        {
            classesHaveChanged();
        }
        private void classesHaveChanged()
        {
            if (listener != null)
            {
                listener.classesChanged(path);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean exists(String... path)
    {
        String fullPath = makeFullPath(path);
        return zk.exists(fullPath);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getString(final ConfigDataListener listener, String... path)
    {
        String fullPath = makeFullPath(path);
        String data = null;
        try
        {
            data = zk.readData(fullPath, true);
            // We can subscribe to the path even if it is not there
            IZkDataListener zkListener = new ZkConfigDataListener(listener);
            zk.subscribeDataChanges(fullPath, zkListener);
        }
        catch (ZkNoNodeException e)
        {
            // This is fine.  The contract is that we'll return null
        }
        // Done
        if (logger.isDebugEnabled())
        {
            logger.debug("\n" +
                    "Loaded string data from ZooKeeper: \n" +
                    "   Path:    " + fullPath + "\n" +
                    "   Data:    " + data == null ? null : (data.length() > 1024 ? data.substring(0, 1024) : data));
        }
        return data;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Properties getProperties(final ConfigDataListener listener, String... path)
    {
        String dataStr = getString(listener, path);
        Properties properties = null;
        if (dataStr != null)
        {
            try
            {
                ZkConfigNode node = new ZkConfigNode(dataStr);
                properties = (Properties) node.getInstance();
            }
            catch (Throwable e)
            {
                throw new ConfigException(
                        "Failed to convert data to properties: \n" +
                        "   Data: " + dataStr + "\n" +
                        "   Path: " + makeFullPath(path),
                        e);
            }
        }
        else
        {
            properties = new Properties();
        }
        // Done (listener added by getString)
        if (logger.isDebugEnabled())
        {
            logger.debug("\n" +
                    "Loaded properties from ZooKeeper: \n" +
                    "   Path:       " + makeFullPath(path) + "\n" +
                    "   Properties: " + properties);
        }
        return properties;
    }

    @Override
    public String setProperties(
            ConfigDataListener listener,
            boolean ephemeral, boolean sequential,
            Properties properties,
            String ... path)
    {
        String fullPath = makeFullPath(path);
        String propertiesStr = null;
        if (properties != null)
        {
            // Fix any properties so that they are escaped properly
            Properties escapedProperties = new Properties();
            for (Map.Entry<Object, Object> entry : properties.entrySet())
            {
                String key = (String) entry.getKey();
                String value = (String) entry.getValue();
                String escapedValue = JSONValue.escape(value);
                if (!value.equals(escapedValue))
                {
                    value = "'" + value + "'";
                }
                escapedProperties.setProperty(key, value);
            }
            propertiesStr = escapedProperties.toString();
        }
        String pathCreated = fullPath;
        if (ephemeral)
        {
            if (sequential)
            {
                pathCreated = zk.createEphemeralSequential(fullPath, propertiesStr);
            }
            else
            {
                zk.createEphemeral(fullPath, propertiesStr);
            }
        }
        else
        {
            if (sequential)
            {
                pathCreated = zk.createEphemeralSequential(fullPath, propertiesStr);
            }
            else
            {
                zk.createPersistent(fullPath, propertiesStr);
            }
        }
        // Subscribe to the path
        IZkDataListener zkListener = new ZkConfigDataListener(listener);
        zk.subscribeDataChanges(pathCreated, zkListener);
        // Extract the last path element
        int index = pathCreated.lastIndexOf(SEPARATOR);
        if (index < 0)
        {
            return pathCreated;
        }
        else
        {
            return pathCreated.substring(index + 1);
        }
    }

    @Override
    public Set<String> getChildren(ConfigChildListener listener, String... path)
    {
        String fullPath = makeFullPath(path);
        Set<String> children = null;
        try
        {
            List<String> childrenList = zk.getChildren(fullPath);
            children = new HashSet<String>(childrenList);
            // Add a listener
            IZkChildListener zkListener = new ZkConfigChildListener(listener);
            zk.subscribeChildChanges(fullPath, zkListener);
        }
        catch (ZkNoNodeException e)
        {
            // OK to return empty set
            children = Collections.emptySet();
        }
        // Done
        if (logger.isDebugEnabled())
        {
            logger.debug("\n" +
                    "Loaded children from ZooKeeper: \n" +
                    "   Path:       " + fullPath + "\n" +
                    "   Children:   " + (children.size() < 11 ? children : "" + children.size()));
        }
        return children;
    }

    @Override
    public ClassLoader getClassLoader(ConfigClassesListener listener, ClassLoader parent, String... path)
    {
        String fullPath = makeFullPath(path);
        ZooKeeperClassLoader zkClassLoader = new ZooKeeperClassLoader(parent, zk, fullPath);
        // Register a listener
        ZkConfigClassLoaderListener zkConfigClassLoaderListener = new ZkConfigClassLoaderListener(fullPath, listener);
        zkClassLoader.addListener(zkConfigClassLoaderListener);
        // Done
        if (logger.isDebugEnabled())
        {
            logger.debug("\n" +
                    "Loaded ZooKeeper ClassLoader: \n" +
                    "   Path:        " + fullPath + "\n" +
                    "   ClassLoader: " + zkClassLoader);
        }
        return zkClassLoader;
    }

    @Override
    public ClassLoader getJarLoader(ConfigClassesListener listener, ClassLoader parent, String... path)
    {
        String fullPath = makeFullPath(path);
        Set<String> children = null;
        try
        {
            List<String> childrenList = zk.getChildren(fullPath);
            children = new HashSet<String>(childrenList);
            // Add a listener
            IZkChildListener zkListener = new ZkConfigJarLoaderListener(fullPath, listener);
            zk.subscribeChildChanges(fullPath, zkListener);
        }
        catch (ZkNoNodeException e)
        {
            // OK to return empty set
            children = Collections.emptySet();
        }
        // Now pull out all the jars
        ArrayList<URL> jarUrlSet = new ArrayList<URL>(5);
        for (String jarStr : children)
        {
            String jarPath = fullPath + SEPARATOR + jarStr;
            if (!jarPath.endsWith(".jar"))
            {
                logger.warn("Found jar entry that is probably not a jar file: " + jarPath);
                continue;
            }
            try
            {
                byte[] jarBytes = zk.readData(jarPath);
                // Write to file
                File jarFile = File.createTempFile("ZkConfigService-JarLoader-", ".jar");
                jarFile.deleteOnExit();
                FileUtils.writeByteArrayToFile(jarFile, jarBytes, false);
                URL jarUrl = jarFile.toURI().toURL();
                jarUrlSet.add(jarUrl);
            }
            catch (Throwable e)
            {
                logger.error("Failed to load jar file: " + jarPath, e);
            }
        }
        URL[] jarUrls = new URL[jarUrlSet.size()];
        jarUrls = jarUrlSet.toArray(jarUrls);
        // Make the classloader
        ClassLoader cl = parent;
        try
        {
            cl = new URLClassLoader(jarUrls, parent);
        }
        catch (Throwable e)
        {
            logger.error("Failed to construct classloader using jars: " + jarUrls, e);
        }
        // Done
        if (logger.isDebugEnabled())
        {
            logger.debug("\n" +
                    "Created URL classloader: \n" +
                    "   Path:       " + fullPath + "\n" +
                    "   Children:   " + (children.size() < 11 ? children : "" + children.size()) + "\n" +
                    "   URLs:       " + jarUrls);
        }
        return cl;
    }
}
