/*
 * Copyright (C) 2005-2012 Alfresco Software Limited.
 *
 * This file is part of Alfresco
 *
 * Alfresco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Alfresco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Alfresco. If not, see <http://www.gnu.org/licenses/>.
 */
package org.alfresco.zookeeper;

import java.io.ByteArrayInputStream;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;

import org.I0Itec.zkclient.exception.ZkMarshallingError;
import org.I0Itec.zkclient.serialize.ZkSerializer;
import org.apache.tika.metadata.Metadata;
import org.apache.tika.parser.AutoDetectParser;
import org.apache.tika.parser.ParseContext;
import org.apache.tika.parser.Parser;
import org.xml.sax.ContentHandler;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A {@link ZkSerializer serializer} that is compatible with the 
 * ZooKeeper command line client (<b>ZkCli</b>).
 * <p/>
 * TODO: Handle different encodings because the <b>ZkCli</b> does not!
 * 
 * @author Derek Hulley
 * @since 1.0
 */
public class ZkStringSerializer implements ZkSerializer
{
    /**
     * @param data          String data
     * @return              the <b>UTF-8</b> byte representation of the data
     */
    @Override
    public byte[] serialize(Object data) throws ZkMarshallingError
    {
        try
        {
            if (data == null)
            {
                throw new ZkMarshallingError("Unable to serialize nulls.");
            }
            if (data instanceof byte[])
            {
                // Just write it out
                return (byte[]) data;
            }
            else
            {
                String dataStr = (String) data;
                byte[] utf8Bytes = dataStr.getBytes("UTF-8");
                // Done
                return utf8Bytes;
            }
        }
        catch (ClassCastException e)
        {
            throw new ZkMarshallingError("Data must be a JSON string", e);
        }
        catch (UnsupportedEncodingException e)
        {
            throw new ZkMarshallingError("Unable to convert String data to UTF8 bytes.", e);
        }
    }

    /**
     * TODO: Cater for mixed charsets inboud
     * 
     * @param bytes         the <b>UTF-8</b> byte representation of a String.
     */
    @Override
    public Object deserialize(byte[] bytes) throws ZkMarshallingError
    {
        ByteArrayInputStream is = new ByteArrayInputStream(bytes);
        try
        {
            // Use Tika to determine what type it is
            Parser parser = new AutoDetectParser();
            ContentHandler handler = new DefaultHandler();
            Metadata metadata = new Metadata();
            ParseContext parseContext = new ParseContext();
            parser.parse(is, handler, metadata, parseContext);
            // Do we have text?
            String mimetype = metadata.get("Content-Type");
            if (mimetype == null)
            {
                // No clue.  Go with a string.
            }
            else if (!mimetype.startsWith("text") && !mimetype.equals("application/xml"))
            {
                // It is not text as far as Tika is concerned
                return bytes;
            }
            // TODO: Fix assumption: "UTF-8"
            String dataStr = new String(bytes, Charset.forName("UTF-8"));
            
            // First check if it's not perhaps plain XML rather than JSon
            if (dataStr.contains("<?xml") || dataStr.contains("<beans"))
            {
                // It is XML.  Just return it.
                return dataStr;
            }
            
            // Trim it
            dataStr = dataStr.trim();
            // Remove quotes
            int startQuoteIndex = dataStr.indexOf('\"');
            if (startQuoteIndex > -1)
            {
                // There must be an end quote
                int endQuoteIndex = dataStr.lastIndexOf('\"');
                if (endQuoteIndex <= startQuoteIndex)
                {
                    throw new ZkMarshallingError("Failed to find end quote after start quote in string: " + dataStr);
                }
                dataStr = dataStr.substring(startQuoteIndex + 1, endQuoteIndex);
            }
            // Done
            return dataStr;
        }
        catch (Throwable e)
        {
            throw new ZkMarshallingError("Failed to convert bytes to String.", e);
        }
        finally
        {
            try { is.close(); } catch (Throwable e) {}
        }
    }
}
